package de.fhhannover.inform.trust.ifmapj.channel;

import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapException;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.messages.PublishRequest;
import de.fhhannover.inform.trust.ifmapj.messages.SearchRequest;
import de.fhhannover.inform.trust.ifmapj.messages.SearchResult;
import de.fhhannover.inform.trust.ifmapj.messages.SubscribeRequest;

/**
 * Interface describing the Synchronous Send Receive Channel (SSRC) methods.
 * 
 * @author aw
 *
 */
public interface SSRC extends IfmapChannel {
	
	/**
	 * Send a newSession request to MAPS
	 * 
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public void newSession() throws IfmapErrorResult, IfmapException;
	
	/**
	 * Send a newSession request to MAPS with the given parameter
	 * 
	 * @param maxPollResSize the maximum size of poll results that can be
	 * processed by the MAPC given in bytes
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public void newSession(Integer maxPollResSize) throws IfmapErrorResult, IfmapException;
	
	/**
	 * Send endSession request to MAPS
	 * 
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public void endSession() throws IfmapErrorResult, IfmapException;
	
	/**
	 * Send renewSession request to MAPS
	 * 
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public void renewSession() throws IfmapErrorResult, IfmapException;
	
	/**
	 * Send purgePublisher request to MAPS in order to purge the MAPCs
	 * metadata
	 * 
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public void purgePublisher() throws IfmapErrorResult, IfmapException;
	
	/**
	 * Send purgePublisher request with the given parameter to MAPS
	 * 
	 * @param publisherId the publisher-id of the MAPC whose metadata should be purged
	 * 
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public void purgePublisher(String publisherId) throws IfmapErrorResult, IfmapException;
	
	/**
	 * Send a publish request to MAPS.
	 * 
	 * @param req the publish request that is sent
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public void publish(PublishRequest req) throws IfmapErrorResult, IfmapException;
	
	/**
	 * Send a subscribe request to the MAPS.
	 * 
	 * @param req the subscribe request
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public void subscribe(SubscribeRequest req) throws IfmapErrorResult, IfmapException;
	
	/**
	 * Send a search request to the MAPS
	 * 
	 * @param req the search request
	 * @return the results of the search
	 * @throws IfmapErrorResult
	 * @throws IfmapException
	 */
	public SearchResult search(SearchRequest req) throws IfmapErrorResult, IfmapException;

	/* public void dump(...) throws IOException, IfmapError; */
	
	/**
	 * Open a {@link ARC} corresponding to this {@link SSRC} instance
	 * 
	 * @return the new {@link ARC}
	 */
	public ARC getArc() throws InitializationException;
}
